/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view;

import java.util.Collection;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.jface.viewers.ViewerFilter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Text;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ElementNameEditingSupport;
import org.fortiss.tooling.base.ui.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;

/**
 * <p>
 * This class provides a generic annotation view whose methods are intended to be reused by child
 * classes. Inherited classes can avoid code duplication, e.g. the creation of the table and its
 * first row, using this class.
 * </p>
 * <p>
 * In order to support displaying and editing specific annotation types, the corresponding
 * {@link IAnnotationValueProvider#createEditingSupport(org.eclipse.jface.viewers.ColumnViewer, Class, String)}
 * factory method must return an appropriate {@link EditingSupport}.
 * </p>
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: D91763408246ECD3578DAD0E37BC4C41
 */
public class GenericAnnotationView extends AnnotationViewPartBase {

	/** The table viewer */
	protected TableViewer tableViewer;

	/**
	 * Character used to separate the annotation name from the instance name for multi-instance
	 * annotations.
	 */
	private final static String MULTIINSTANCE_NAME_SEPARATOR = ":";

	/** Widget for editing of filter text. */
	private Text filterText;

	/** Filter hint text. */
	private static final String FILTER_HINT_TEXT = "type filter text";

	/** Filter pattern */
	private String filterPattern;

	/** Flag whether columns should be filtered. */
	boolean filterColumns;

	/**
	 * Flag whether the displayed rows should be restricted to the currently selected model element
	 * type.
	 */
	boolean restrictToSelectedModelElementType;

	/**
	 * Last set of {@link AnnotationEntry}s used to {@link #update(Collection)} the
	 * {@link GenericAnnotationView}, i.e. during the construction of the, or because of a model
	 * change. The reference is required to update view after a change of the column (annotation)
	 * filter.
	 */
	Collection<AnnotationEntry> lastAnnotationEntries;

	/**
	 * Data required to identify a column displaying a particular {@link IAnnotatedSpecification} in
	 * a column of the {@link GenericAnnotationView}. Used to sort columns (see
	 * {@link #compareTo(ColumnHandle)}).
	 */
	private class ColumnHandle implements Comparable<ColumnHandle> {
		/**
		 * {@link AnnotationEntry} from which a particular {@link IAnnotatedSpecification} is to be
		 * displayed in the column represented by this {@link ColumnHandle}.
		 */
		private AnnotationEntry entry;

		/**
		 * {@link IAnnotatedSpecification} from {@code entry} to be displayed in the column
		 * represented by this {@link ColumnHandle}.
		 */
		private IAnnotatedSpecification annotatedSpecification;

		/**
		 * Instance key identifying the particular instance of an {@link IAnnotatedSpecification}
		 * from {@code entry} to be displayed in the column represented by this {@link ColumnHandle}
		 * .
		 */
		private String instanceKey;

		/** Flag if this {@link ColumnHandle} represents a {@link CreateAnnotationInstanceColumn}. */
		private boolean isCreateColumn;

		/** Constructs a new {@link ColumnHandle}. */
		public ColumnHandle(AnnotationEntry entry, IAnnotatedSpecification annotatedSpecification,
				String instanceKey, boolean isCreateColumn) {

			this.entry = entry;
			this.annotatedSpecification = annotatedSpecification;
			this.instanceKey = instanceKey;
			this.isCreateColumn = isCreateColumn;
		}

		/**
		 * Returns the {@link AnnotationEntry} from which a particular
		 * {@link IAnnotatedSpecification} is to be
		 * displayed in the column represented by this {@link ColumnHandle}.
		 */
		public AnnotationEntry getEntry() {
			return entry;
		}

		/**
		 * Returns the {@link IAnnotatedSpecification} from {@code entry} to be displayed in the
		 * column represented by this {@link ColumnHandle}.
		 */
		public IAnnotatedSpecification getAnnotatedSpecification() {
			return annotatedSpecification;
		}

		/**
		 * Returns the instance key identifying the particular instance of an
		 * {@link IAnnotatedSpecification} from {@code entry} to be displayed in the column
		 * represented by this {@link ColumnHandle} .
		 */
		public String getInstanceKey() {
			return instanceKey;
		}

		/**
		 * Returns a flag if this {@link ColumnHandle} represents a
		 * {@link CreateAnnotationInstanceColumn}.
		 */
		public boolean isCreateColumn() {
			return isCreateColumn;
		}

		/** {@inheritDoc} */
		@Override
		public int compareTo(ColumnHandle other) {
			String columnName = getColumnName(this);
			String annotationName;

			int columnNameSeparatorIndex = columnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
			if(columnNameSeparatorIndex != -1) {
				annotationName = columnName.substring(0, columnNameSeparatorIndex);
			} else {
				annotationName = columnName;
			}

			String otherColumnName = getColumnName(other);
			String otherAnnotationName;
			int otherColumnNameSeparatorIndex =
					otherColumnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
			if(otherColumnNameSeparatorIndex != -1) {
				otherAnnotationName = otherColumnName.substring(0, otherColumnNameSeparatorIndex);
			} else {
				otherAnnotationName = otherColumnName;
			}

			// Place "create" columns (label: "<specification name>") after their instances
			// (which are labeled "<specification name>: <instance name>").
			if(annotationName.equals(otherAnnotationName)) {
				// Ensure uniqueness of "create" columns"
				if(isCreateColumn && other.isCreateColumn) {
					return 0;
				}
				if(isCreateColumn) {
					return 1;
				}
				if(other.isCreateColumn) {
					return -1;
				}
			}

			// Sort alphabetically by specification column label
			return columnName.compareTo(otherColumnName);
		}
	}

	/**
	 * Comparator that sorts rows according to hierarchy level. Elements within one hierarchy level
	 * are sorted by name.
	 */
	private class HierarchicalNameComparator extends ViewerComparator {

		/** Computes the level of an {@link IModelElement} in the tree */
		private int getLevel(EObject o) {
			int level = 0;
			while(o.eContainer() != null) {
				level++;
				o = o.eContainer();
			}

			return level;
		}

		/**
		 * Returns the ancestor of the {@link IModelElement} {@code modelElement} that is
		 * {@code levelsUp} levels above model element in the hierarchy, or {@code null}, in case
		 * {@code levelsUp} > depth(modelElement}.
		 */
		private EObject getAncestor(EObject o, int levelsUp) {
			while(o.eContainer() != null && levelsUp > 0) {
				o = o.eContainer();
				levelsUp--;
			}
			return o;

		}

		/** {@inheritDoc} */
		@Override
		public int compare(Viewer viewer, Object o1, Object o2) {
			if(o1 instanceof AnnotationEntry && o2 instanceof AnnotationEntry) {
				IModelElement modelElement1 = ((AnnotationEntry)o1).getModelElement();
				IModelElement modelElement2 = ((AnnotationEntry)o2).getModelElement();
				int modelElement1Level = getLevel(modelElement1);
				int modelElement2Level = getLevel(modelElement2);

				EObject e1 = modelElement1;
				EObject e2 = modelElement2;

				if(modelElement1Level < modelElement2Level) {
					// Element1 is higher than element2: Get ancestor of element2 that is at the
					// level as the element1.
					EObject ancestor =
							getAncestor(modelElement2, modelElement2Level - modelElement1Level);

					if(modelElement1 == ancestor) {
						// Element1 is offspring of element2
						return -1;
					}

					// Further comparison starts on elements at the same level
					e2 = ancestor;
				} else if(modelElement1Level > modelElement2Level) {
					// Element2 is higher than element1: Get ancestor of element1 that is at the
					// level as the element2.
					EObject ancestor =
							getAncestor(modelElement1, modelElement1Level - modelElement2Level);

					if(modelElement2 == ancestor) {
						// Element2 is offspring of element1
						return 1;
					}

					// Further comparison starts on elements at the same level
					e1 = ancestor;
				}
				// Find direct children of common root element of element1 and element2...
				do {
					if(e1.eContainer() == e2.eContainer()) {
						break;
					}
					e1 = e1.eContainer();
					e2 = e2.eContainer();
				} while(true);

				// ... and compare them by name...
				if(e1 instanceof INamedElement && e2 instanceof INamedElement) {
					// Elements are at the same level -> compare by name
					return(((INamedElement)e1).getName().compareTo(((INamedElement)e2).getName()));
				}
			}

			// ... or hash code (as a fallback ordering)
			// int Integer.compare(int x, int y) requires JDK 1.7 which is not available
			// on the build server
			return new Integer(o1.hashCode()).compareTo(o2.hashCode());
		}
	}

	// Based on org.fortiss.af3.scheduling.ui.specificationsEditor.ComponentResourceTableBackend
	/** Filters rows of the {@link GenericAnnotationView} based on the components' names. */
	private class ComponentNamesRowFilter extends ViewerFilter {
		/** Flag if the filter is active */
		private boolean active = true;

		/** (De-)activates the filter */
		public void setActive(boolean active) {
			this.active = active;
		}

		/** {@inheritDoc} */
		@Override
		public boolean select(Viewer viewer, Object parentElement, Object element) {
			if(!(element instanceof AnnotationEntry)) {
				return true;
			}

			AnnotationEntry annotationEntry = ((AnnotationEntry)element);
			if(!(annotationEntry.getModelElement() instanceof INamedElement)) {
				return true;
			}

			return (!active || passesFilter(
					((INamedElement)annotationEntry.getModelElement()).getName(), filterPattern)) &&
					(!restrictToSelectedModelElementType || getCurrentlySelectedObject() == null ||
							getCurrentlySelectedObject() instanceof IProjectRootElement || annotationEntry
							.getModelElement().getClass()
							.equals(getCurrentlySelectedObject().getClass()));
		}
	}

	/**
	 * Returns {@code true} if a given {@code input} passes an case-insensitive filter specified by
	 * {@code filterString}.
	 */
	private static boolean passesFilter(String input, String filterString) {
		// Null-filter accepts every input
		if(filterString == null || filterString.equals(FILTER_HINT_TEXT)) {
			return true;
		}

		// Null-input cannot be accepted by a non-null filter
		if(input == null) {
			return false;
		}
		return input.toLowerCase().contains(filterString.toLowerCase());
	}

	/** Returns {@code true} if the column represented by the given column is visible in the view. */
	private boolean passesFilter(ColumnHandle columnHandle, String filterPattern) {
		if(!filterColumns) {
			return true;
		}
		return passesFilter(getColumnName(columnHandle), filterPattern);
	}

	/** Updates the table after a filter change */
	private void update() {
		// Full update, i.e. including reconstructions of columns is always required due to the
		// following scenario:
		// - Filter is set for annotation names.
		// - Filter scope is changed to component names -> annotation filter needs to be undone
		if(lastAnnotationEntries != null) {
			update(lastAnnotationEntries);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void update(Collection<AnnotationEntry> annotationEntries) {
		// Remember last annotation entries (required to re-draw table after changing the
		// column/annotation filter
		lastAnnotationEntries = annotationEntries;

		// Minimize flickering while updating the table
		tableViewer.getTable().setRedraw(false);

		// Input must be set before the construction of the columns, since the editing support for a
		// given cell might need to query the contents of other model elements (e.g., in order to
		// aggregate a set of admissible choices for its value).
		tableViewer.setInput(annotationEntries);

		/*
		 * Sorted set of {@link ColumnHandle} used to instantiate this {@link
		 * GenericAnnotationView}'s columns in the right order.
		 */
		Set<ColumnHandle> sortedColumnHandles = new TreeSet<ColumnHandle>();

		/*
		 * Dispose all columns of the table viewer, except the first column, which holds the names
		 * of the model elements
		 */
		while(tableViewer.getTable().getColumnCount() > 1) {
			tableViewer.getTable().getColumn(1).dispose();
		}

		// Aggregate required columns. Column order is defined by ColumnHandle.compareTo().
		for(AnnotationEntry entry : annotationEntries) {
			for(IAnnotatedSpecification spec : entry.getSpecificationsList()) {
				// Create new column for each instance of the current multi-instance annotation
				if((entry.allowsDynamicAnnotationInstances(spec.getClass())) ||
						!entry.getInstanceKeys(spec.getClass()).isEmpty()) {

					for(String instanceKey : entry.getInstanceKeys(spec.getClass())) {
						ColumnHandle columnHandle =
								new ColumnHandle(entry, spec, instanceKey, false);
						if(passesFilter(columnHandle, filterPattern)) {
							sortedColumnHandles.add(columnHandle);
						}
					}
				}

				// Create a new column for
				// - ordinary (single-instance) annotations
				// - "create" column for multi-instance annotations for which instances can be
				// created dynamically
				if(entry.getInstanceKeys(spec.getClass()).isEmpty() ||
						entry.allowsDynamicAnnotationInstances(spec.getClass())) {
					ColumnHandle columnHandle =
							new ColumnHandle(entry, spec, null,
									entry.allowsDynamicAnnotationInstances(spec.getClass()));

					if(passesFilter(columnHandle, filterPattern)) {
						sortedColumnHandles.add(columnHandle);
					}

				}
			}
		}

		// Instantiate columns
		for(ColumnHandle columnHandle : sortedColumnHandles) {
			createAnnotationColumn(columnHandle);
		}

		// Update the view
		tableViewer.getTable().setRedraw(true);
		tableViewer.refresh();
	}

	/** Creates the UI controls for the column and row filter. */
	private void createFilter(Composite rootComposite) {
		// Rows (= components) are filtered using a specialized ViewerFilter.
		// Since the data model is an array of AnnotationEntries, columns (= annotations)
		// cannot be filtered using such a filter, and require the reconstruction of the
		// table, with an implementation of the filter in update(Set<AnnotationEntry>).
		final ComponentNamesRowFilter rowfilter = new ComponentNamesRowFilter();
		rowfilter.setActive(true);
		tableViewer.addFilter(rowfilter);

		// Composite for filter GUI controls
		Composite filterComposite = new Composite(rootComposite, SWT.NULL);
		filterComposite.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		filterComposite.setLayout(new GridLayout(2, false));

		// Text input field for filter pattern
		filterText =
				new Text(filterComposite, SWT.SINGLE | SWT.BORDER | SWT.SEARCH | SWT.ICON_CANCEL);
		filterText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, true, 1, 1));
		filterText.setText(FILTER_HINT_TEXT);
		filterText.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GRAY));

		// Observe changes of filter pattern
		filterText.addModifyListener(new ModifyListener() {
			@Override
			public void modifyText(ModifyEvent e) {
				filterPattern = filterText.getText();
				if(filterPattern.isEmpty()) {
					filterText.setText(FILTER_HINT_TEXT);
					filterText.selectAll();
					filterPattern = FILTER_HINT_TEXT;
				}
				if(filterPattern.equals(FILTER_HINT_TEXT)) {
					filterText.setForeground(Display.getCurrent().getSystemColor(
							SWT.COLOR_DARK_GRAY));
				} else {
					filterText.setForeground(Display.getCurrent().getSystemColor(
							SWT.COLOR_WIDGET_FOREGROUND));
				}
				update();
			}
		});

		// Select the filter pattern if it is equal to the FILTER_HINT_TEXT
		TextFocusListener textSelectionOnFocusListener = new TextFocusListener() {

			@Override
			protected void focusOut(Text text, Event e) {
				text.clearSelection();

			}

			@Override
			protected void focusIn(Text text, Event e) {
				if(text.getText().equals(FILTER_HINT_TEXT) &&
						(e.type != SWT.MouseUp || text.getSelectionCount() == 0)) {
					text.selectAll();
				}

			}
		};
		filterText.addListener(SWT.FocusIn, textSelectionOnFocusListener);
		filterText.addListener(SWT.FocusOut, textSelectionOnFocusListener);
		filterText.addListener(SWT.MouseDown, textSelectionOnFocusListener);
		filterText.addListener(SWT.MouseUp, textSelectionOnFocusListener);

		// Filter options
		Composite filterOptionsComposite = new Composite(filterComposite, SWT.NULL);
		filterOptionsComposite
				.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1));
		filterOptionsComposite.setLayout(new RowLayout());

		// Radio buttons to select between row and column filtering
		final Label filterLabel = new Label(filterOptionsComposite, SWT.NULL);
		filterLabel.setText("Filter ");

		final Button filterComponentNamesButton = new Button(filterOptionsComposite, SWT.RADIO);
		filterComponentNamesButton.setText("component");
		filterComponentNamesButton.setSelection(true);
		filterComponentNamesButton.addSelectionListener(new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				// widgetSelected() is also fired when the selection is removed from the radio
				// button, so check, if this is the button that just obtained the selection
				if(!((Button)e.getSource()).getSelection()) {
					return;
				}
				rowfilter.setActive(true);
				filterColumns = false;
				update();
			}
		});

		final Button filterAnnotationNamesButton = new Button(filterOptionsComposite, SWT.RADIO);
		filterAnnotationNamesButton.setText("annotation names.");
		filterAnnotationNamesButton.setSelection(false);
		filterAnnotationNamesButton.addSelectionListener(new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				// widgetSelected() is also fired when the selection is removed from the radio
				// button, so check, if this is the button that just obtained the selection
				if(!((Button)e.getSource()).getSelection()) {
					return;
				}
				rowfilter.setActive(false);
				filterColumns = true;
				update();
			}
		});

		final Label spacerLabel = new Label(filterOptionsComposite, SWT.NULL);
		spacerLabel.setText("  ");

		// Check box to restrict view to currently selected model element type
		final Button filterModelElementType = new Button(filterOptionsComposite, SWT.CHECK);
		filterModelElementType.setText("Show only selected model element type.");
		filterModelElementType.setSelection(false);
		filterModelElementType.addSelectionListener(new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				// widgetSelected() is also fired when the selection is removed from the radio
				// button, so check, if this is the button that just obtained the selection
				restrictToSelectedModelElementType = ((Button)e.getSource()).getSelection();
				update();
			}
		});

	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		super.createPartControl(parent);
		Composite rootComposite = new Composite(parent, SWT.NULL);
		rootComposite.setLayout(new GridLayout(1, false));

		tableViewer = new TableViewer(rootComposite, SWT.BORDER | SWT.FULL_SELECTION);
		Table table = tableViewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);
		table.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));

		createModelElementColumn();

		tableViewer.setComparator(new HierarchicalNameComparator());

		createFilter(rootComposite);
	}

	/** Creates the (leading) column which displays the model elements */
	protected void createModelElementColumn() {

		final TableViewerColumn firstColumn = new TableViewerColumn(tableViewer, SWT.NONE);
		firstColumn.getColumn().setText("IModelElement");
		firstColumn.getColumn().setWidth(100);

		tableViewer.setContentProvider(new ArrayContentProvider());
		firstColumn.setLabelProvider(new ElementNameLabelProvider(this, firstColumn));
		firstColumn.setEditingSupport(new ElementNameEditingSupport(tableViewer));
	}

	/** Creates a column in tableViewer that displays the annotation given in spec */
	protected void createAnnotationColumn(ColumnHandle columnHandle) {

		TableColumn tableColumn = null;
		if(columnHandle.isCreateColumn()) {
			// Create column that contains a button for adding new instances of the
			// current annotation type.

			CreateAnnotationInstanceColumn column =
					new CreateAnnotationInstanceColumn(tableViewer, SWT.NONE,
							columnHandle.getAnnotatedSpecification(), columnHandle.getEntry());
			tableColumn = column.getColumn();

		} else {
			Class<? extends IAnnotatedSpecification> annotationClass =
					columnHandle.getAnnotatedSpecification().getClass();

			// Add new column
			try {

				// Have the matching EditingSupport created for the current annotated specification
				// (work is delegated to respective IAnnotationValueProvider implementation)
				EditingSupport editingSupport =
						columnHandle.getEntry().createSpecificationEditElement(tableViewer,
								annotationClass, columnHandle.getInstanceKey());
				TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);
				tableColumn = column.getColumn();

				String specName = getColumnName(columnHandle);
				tableColumn.setText(specName);

				column.setEditingSupport(editingSupport);

				// Add column label provider, and set EditingSupport
				column.setLabelProvider(new AnnotationLabelProvider(annotationClass, this,
						columnHandle.getInstanceKey()));
			} catch(Exception e) {
				System.out.println("Failed to create editing support for annotation class " +
						annotationClass.getCanonicalName() + ": " + e.getMessage());
			}

		}

		if(tableColumn != null) {
			tableColumn.setWidth(125);
		}

	}

	/**
	 * Returns the name of a column for a given {@link IAnnotatedSpecification} represented by the
	 * given {@link ColumnHandle}.
	 */
	protected String getColumnName(ColumnHandle columnHandle) {
		// Determine column name
		String specName =
				columnHandle.getEntry().getSpecificationAnnotationName(
						columnHandle.getAnnotatedSpecification().getClass());

		if(columnHandle.getInstanceKey() != null) {
			if(!specName.isEmpty()) {
				specName += MULTIINSTANCE_NAME_SEPARATOR + " ";
			}
			specName += columnHandle.getInstanceKey();
		} else if(specName == null || specName.isEmpty()) {
			specName = "<Unnamed Annotation>";
		}
		return specName;
	}

	/** {@inheritDoc} */
	@Override
	public void setFocus() {
		// nothing to do
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		tableViewer.getTable().dispose();
		super.dispose();
	}

}
